# Copyright (c) 2025 Red Hat Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(ralonsoh): remove once the default backend is ``BackendType.THREADING``
from oslo_service import backend as oslo_service_backend
oslo_service_backend.init_backend(oslo_service_backend.BackendType.THREADING)

# pylint: disable=wrong-import-position
from neutron import server  # noqa: E402
from neutron.server import api  # noqa: E402
from neutron.server import ovn_maintenance  # noqa: E402
from neutron.server import periodic  # noqa: E402
from neutron.server import rpc  # noqa: E402


def main_api_uwsgi():
    return server.boot_server(api.api_server)


def main_ovn_maintenance():
    return server.boot_server(ovn_maintenance.ovn_maintenance_worker)


def main_rpc():
    server.boot_server(rpc.rpc_server)


def main_periodic():
    server.boot_server(periodic.periodic_workers)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import netaddr
from oslo_log import log

from neutron.services.bgp import constants

LOG = log.getLogger(__name__)

BGP_LRP_UUID_NAMESPACE = uuid.UUID('9eaaac59-33fc-4f45-a450-8c220d46ad95')


def get_mac_address_from_lrp_name(lrp_name):
    mac_uuid_base = uuid.uuid5(BGP_LRP_UUID_NAMESPACE, lrp_name)

    # Let's take last 6 bytes of the UUID and convert them to bytes
    mac_bytes = bytearray(mac_uuid_base.bytes[-6:])

    # 3. Apply Bitwise Operations on the First Byte
    # Set the "Locally Administered" bit (2nd least significant bit) to 1
    # xxxxxxx1 | 00000010 = xxxxxxx1 (OR 0x02)
    mac_bytes[0] |= 0x02

    # Clear the "Multicast" bit (Least significant bit) to 0
    # xxxxxxx1 & 11111110 = xxxxxxx0 (AND 0xfe)
    mac_bytes[0] &= 0xfe

    # 4. Format into standard MAC string (XX:XX:XX:XX:XX:XX)
    mac_address = ':'.join(f'{b:02x}' for b in mac_bytes)

    return mac_address


def get_chassis_bgp_bridges(chassis):
    try:
        bgp_bridges = chassis.external_ids[
            constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY]
    except KeyError:
        LOG.warning("Chassis %s has no BGP bridges set", chassis.name)
        return []

    return [bridge for br in bgp_bridges.split(',') if (bridge := br.strip())]


# Naming helper functions
def get_lrp_name(from_name, to_name):
    return f'bgp-lrp-{from_name}-to-{to_name}'


def get_lsp_name(from_name, to_name):
    return f'bgp-lsp-{from_name}-to-{to_name}'


def get_lsp_localnet_name(switch_name):
    return f'bgp-lsp-{switch_name}-localnet'


def get_hcg_name(chassis_name):
    return f'bgp-hcg-{chassis_name}'


def get_chassis_router_name(chassis_name):
    return f'bgp-lr-{chassis_name}'


def get_chassis_peer_switch_name(chassis_name, network_name):
    return f'bgp-ls-{chassis_name}-{network_name}'


def get_provider_interconnect_switch_name(provider_switch_name):
    return f'bgp-ls-interconnect-{provider_switch_name}'


def get_ip_network(ip_address, ip_network):
    return f"{ip_address}/{ip_network.prefixlen}"


def get_neutron_id_from_ovn_name(ovn_obj):
    # Get the ID from the OVN name, which is not ideal but the
    # network ID does not seem to be stored elsewhere in the OVN objects.
    try:
        return ovn_obj.name.split('neutron-', 1)[1]
    except IndexError:
        raise ValueError(
            f"OVN object {ovn_obj.name} does not contain a Neutron ID")


## Router filtering helpers
def _get_lrps_by_external_id(router, external_id):
    return [lrp for lrp in router.ports
            if hasattr(lrp, 'external_ids') and
            external_id in lrp.external_ids]


def lrps_to_chassis_routers(router):
    return _get_lrps_by_external_id(router, constants.BGP_LRP_TO_CHASSIS)


def ipv6_link_local_from_mac(mac):
    return str(netaddr.EUI(mac).ipv6_link_local())


def get_gw_ip_from_dhcp_options(dhcp_opt):
    prefixlen = netaddr.IPNetwork(dhcp_opt.cidr).prefixlen
    try:
        return f"{dhcp_opt.options['router']}/{prefixlen}"
    except KeyError:
        return None



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import signal
import threading
import time
import unittest
from unittest import mock

from oslo_utils import importutils


MODULE = ('neutron.plugins.ml2.drivers.openvswitch.agent.'
          'openflow.native.ovs_oskenapp')


class TestSignalHandling(unittest.TestCase):

    def setUp(self):
        super().setUp()

        os.environ['OSKEN_HUB_TYPE'] = 'native'
        self.ovs_oskenapp = importutils.import_module(MODULE)

    @mock.patch('neutron.plugins.ml2.drivers.openvswitch.agent.openflow.'
                'native.ovs_oskenapp.ovs_agent.main')
    def test_signal_execution_in_thread(self, mock_ovs_agent_main):
        # TODO(ralonsoh): refactor this test to make it compatible after the
        # eventlet removal.
        self.skipTest('This test is skipped after the eventlet removal and '
                      'needs to be refactored')
        # The event is used to validate the handler stop_running() has
        # been called and to synchronize the test
        stop_event = threading.Event()

        def mock_ovs_agent_main_impl(bridge_classes, register_signal):
            running = True

            def stop_running():
                nonlocal running

                running = False
                stop_event.set()

            register_signal(signal.SIGTERM, stop_running)

            while running:
                # Simulate processing by agent.
                time.sleep(0.1)

        mock_ovs_agent_main.side_effect = mock_ovs_agent_main_impl

        app = self.ovs_oskenapp.OVSNeutronAgentOSKenApp()

        self.addCleanup(app.stop)
        app.start()

        # Fire SIGTERM
        os.kill(os.getpid(), signal.SIGTERM)

        stop_event.wait(timeout=2)

        mock_ovs_agent_main.assert_called_once()



# Copyright (c) 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from neutron_lib.api import attributes
from neutron_lib.api.definitions import subnet as subnet_def
from neutron_lib.api.definitions import subnet_external_network as sen_def
from oslo_policy import policy as base_policy
from oslo_utils import uuidutils

from neutron import policy
from neutron.tests.unit.conf.policies import test_base as base


class SubnetAPITestCase(base.PolicyBaseTestCase):

    def setUp(self):
        # Extend subnet "subnet-external-network" extension. This extension
        # is not loaded in the unit tests.
        rname = subnet_def.COLLECTION_NAME
        attributes.RESOURCES[rname].update(
            sen_def.RESOURCE_ATTRIBUTE_MAP[rname])
        super().setUp()

        self.network = {
            'id': uuidutils.generate_uuid(),
            'project_id': self.project_id}
        self.alt_network = {
            'id': uuidutils.generate_uuid(),
            'project_id': self.alt_project_id}
        self.ext_alt_network = {
            'id': uuidutils.generate_uuid(),
            'project_id': self.alt_project_id}

        networks = {
            self.network['id']: self.network,
            self.alt_network['id']: self.alt_network,
            self.ext_alt_network['id']: self.ext_alt_network,
        }

        self.target = {
            'project_id': self.project_id,
            'network_id': self.network['id'],
            'ext_parent_network_id': self.network['id']}
        # This subnet belongs to "project_id", but not the network that
        # belongs to "alt_project_id".
        self.target_net_alt_target = {
            'project_id': self.project_id,
            'network_id': self.alt_network['id'],
            'ext_parent_network_id': self.alt_network['id']}
        self.alt_target = {
            'project_id': self.alt_project_id,
            'network_id': self.alt_network['id'],
            'ext_parent_network_id': self.alt_network['id']}
        # Both the subnet and the network belongs to "alt_project_id" and the
        # network is external.
        self.target_net_ext_alt_target = {
            'project_id': self.alt_project_id,
            'network_id': self.ext_alt_network['id'],
            'ext_parent_network_id': self.ext_alt_network['id'],
            'router:external': True}
        # This is the case where the network belongs to the project but not
        # the subnet.
        self.alt_target_own_net = {
            'project_id': self.alt_project_id,
            'network_id': self.network['id'],
            'ext_parent_network_id': self.network['id']}

        def get_network(context, id, fields=None):
            return networks.get(id)

        self.plugin_mock = mock.Mock()
        self.plugin_mock.get_network.side_effect = get_network
        mock.patch(
            'neutron_lib.plugins.directory.get_plugin',
            return_value=self.plugin_mock).start()


class SystemAdminTests(SubnetAPITestCase):

    def setUp(self):
        super().setUp()
        self.context = self.system_admin_ctx

    def test_create_subnet(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet', self.alt_target_own_net)

    def test_create_subnet_segment_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:segment_id', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:segment_id',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:segment_id', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:segment_id', self.alt_target_own_net)

    def test_create_subnet_service_types(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:service_types', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:service_types',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:service_types', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:service_types',
            self.alt_target_own_net)

    def test_create_subnet_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:tags', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:tags', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_subnet:tags', self.alt_target_own_net)

    def test_get_subnet(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet', self.target_net_ext_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet', self.alt_target_own_net)

    def test_get_subnet_segment_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.alt_target_own_net)

    def test_get_subnet_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:tags', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:tags', self.target_net_ext_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:tags', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_subnet:tags', self.alt_target_own_net)

    def test_update_subnet(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet', self.alt_target_own_net)

    def test_update_subnet_segment_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:segment_id', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:segment_id',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:segment_id', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:segment_id', self.alt_target_own_net)

    def test_update_subnet_service_types(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:service_types', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:service_types',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:service_types', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:service_types',
            self.alt_target_own_net)

    def test_update_subnet_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:tags', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:tags', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_subnet:tags', self.alt_target_own_net)

    def test_delete_subnet(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet', self.alt_target_own_net)

    def test_delete_subnet_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.target_net_alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.alt_target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.alt_target_own_net)


class SystemMemberTests(SystemAdminTests):

    def setUp(self):
        super().setUp()
        self.context = self.system_member_ctx


class SystemReaderTests(SystemMemberTests):

    def setUp(self):
        super().setUp()
        self.context = self.system_reader_ctx


class AdminTests(SubnetAPITestCase):

    def setUp(self):
        super().setUp()
        self.context = self.project_admin_ctx

    def test_create_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet', self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet',
                           self.alt_target_own_net))

    def test_create_subnet_segment_id(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:segment_id', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:segment_id',
                self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:segment_id', self.alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:segment_id',
                self.alt_target_own_net))

    def test_create_subnet_service_types(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:service_types', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:service_types',
                self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:service_types', self.alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_subnet:service_types',
                self.alt_target_own_net))

    def test_create_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet:tags',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet:tags',
                           self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet:tags',
                           self.alt_target_own_net))

    def test_get_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet',
                           self.target_net_ext_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet', self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet',
                           self.alt_target_own_net))

    def test_get_subnet_segment_id(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:segment_id', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:segment_id',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_subnet:segment_id', self.alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_subnet:segment_id',
                self.alt_target_own_net))

    def test_get_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags',
                           self.target_net_ext_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags', self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags',
                           self.alt_target_own_net))

    def test_update_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet', self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet',
                           self.alt_target_own_net))

    def test_update_subnet_segment_id(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_subnet:segment_id', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_subnet:segment_id',
                self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_subnet:segment_id', self.alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_subnet:segment_id',
                self.alt_target_own_net))

    def test_update_subnet_service_types(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_subnet:service_types', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_subnet:service_types',
                self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_subnet:service_types', self.alt_target))

    def test_update_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet:tags',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet:tags',
                           self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet:tags',
                           self.alt_target_own_net))

    def test_delete_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet', self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet',
                           self.alt_target_own_net))

    def test_delete_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet:tags',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet:tags',
                           self.alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet:tags',
                           self.alt_target_own_net))


class ProjectManagerTests(AdminTests):

    def setUp(self):
        super().setUp()
        self.context = self.project_manager_ctx

    def test_create_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet', self.alt_target)
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet',
                           self.alt_target_own_net))

    def test_create_subnet_segment_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:segment_id', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:segment_id',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:segment_id', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:segment_id', self.alt_target_own_net)

    def test_create_subnet_service_types(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:service_types', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:service_types',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:service_types', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:service_types',
            self.alt_target_own_net)

    def test_create_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet:tags',
                           self.target_net_alt_target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:tags', self.alt_target)
        self.assertTrue(
            policy.enforce(self.context, 'create_subnet:tags',
                           self.alt_target_own_net))

    def test_get_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet',
                           self.target_net_ext_alt_target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_subnet', self.alt_target)
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet',
                           self.alt_target_own_net))

    def test_get_subnet_segment_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.alt_target_own_net)

    def test_get_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags',
                           self.target_net_alt_target))
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags',
                           self.target_net_ext_alt_target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_subnet:tags', self.alt_target)
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet:tags',
                           self.alt_target_own_net))

    def test_update_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet',
                           self.target_net_alt_target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet', self.alt_target)
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet',
                           self.alt_target_own_net))

    def test_update_subnet_segment_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:segment_id', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:segment_id',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:segment_id', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:segment_id', self.alt_target_own_net)

    def test_update_subnet_service_types(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:service_types', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:service_types',
            self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:service_types', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:service_types',
            self.alt_target_own_net)

    def test_update_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet:tags',
                           self.target_net_alt_target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:tags', self.alt_target)
        self.assertTrue(
            policy.enforce(self.context, 'update_subnet:tags',
                           self.alt_target_own_net))

    def test_delete_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet',
                           self.target_net_alt_target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet', self.alt_target)
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet',
                           self.alt_target_own_net))

    def test_delete_subnet_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet:tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_subnet:tags',
                           self.target_net_alt_target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.alt_target)
        self.assertTrue(
                policy.enforce(self.context, 'delete_subnet:tags',
                               self.alt_target_own_net))


class ProjectMemberTests(ProjectManagerTests):

    def setUp(self):
        super().setUp()
        self.context = self.project_member_ctx


class ProjectReaderTests(ProjectMemberTests):

    def setUp(self):
        super().setUp()
        self.context = self.project_reader_ctx

    def test_create_subnet(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet', self.alt_target_own_net)

    def test_create_subnet_tags(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:tags', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:tags', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:tags', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:tags', self.alt_target_own_net)

    def test_update_subnet(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet', self.alt_target_own_net)

    def test_update_subnet_tags(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:tags', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:tags', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:tags', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:tags', self.alt_target_own_net)

    def test_delete_subnet(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet', self.alt_target_own_net)

    def test_delete_subnet_tags(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.target_net_alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.alt_target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet:tags', self.alt_target_own_net)


class ServiceRoleTests(SubnetAPITestCase):

    def setUp(self):
        super().setUp()
        self.context = self.service_ctx

    def test_create_subnet(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet', self.target)

    def test_create_subnet_segment_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:segment_id', self.target)

    def test_create_subnet_service_types(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:service_types', self.target)

    def test_create_subnet_tags(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_subnet:tags', self.target)

    def test_get_subnet(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_subnet', self.target))

    def test_get_subnet_segment_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_subnet:segment_id', self.target)

    def test_update_subnet(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet', self.target)

    def test_update_subnet_segment_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:segment_id', self.target)

    def test_update_subnet_service_types(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_subnet:service_types', self.target)

    def test_delete_subnet(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_subnet', self.target)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron.services.bgp import constants
from neutron.services.bgp import helpers
from neutron.tests import base


class GetMacAddressFromLrpNameTestCase(base.BaseTestCase):
    def test_get_mac_address_from_lrp_name_unique(self):
        num_macs = 1000
        macs = set()
        for i in range(num_macs):
            lrp_name = f'test-lrp-{i}'
            mac = helpers.get_mac_address_from_lrp_name(lrp_name)
            try:
                netaddr.EUI(mac, version=48)
            except netaddr.AddrFormatError as e:
                self.fail(f"Invalid MAC address generated: {mac}, error: {e}")
            macs.add(mac)

        self.assertEqual(num_macs, len(macs))

    def test_get_mac_address_from_lrp_name_consistent(self):
        lrp_name = 'test-lrp-1'
        mac1 = helpers.get_mac_address_from_lrp_name(lrp_name)
        mac2 = helpers.get_mac_address_from_lrp_name(lrp_name)
        self.assertEqual(mac1, mac2)


class GetChassisBgpBridgesTestCase(base.BaseTestCase):
    class FakeChassis:
        def __init__(self, name, external_ids):
            self.name = name
            self.external_ids = external_ids

    def test_get_chassis_bgp_bridges(self):
        chassis = self.FakeChassis(
            name='test-chassis',
            external_ids={constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY: 'br1,br2'})
        self.assertEqual(['br1', 'br2'],
                         sorted(helpers.get_chassis_bgp_bridges(chassis)))

    def test_get_chassis_bgp_bridges_no_bridges(self):
        chassis = self.FakeChassis(
            name='test-chassis',
            external_ids={})
        self.assertEqual([], helpers.get_chassis_bgp_bridges(chassis))

    def test_get_chassis_bgp_bridges_empty_bridges(self):
        chassis = self.FakeChassis(
            name='test-chassis',
            external_ids={constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY: ''})
        self.assertEqual([], helpers.get_chassis_bgp_bridges(chassis))



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# NOTE(ralonsoh): remove once the default backend is ``BackendType.THREADING``
from oslo_service import backend as oslo_service_backend
import warnings
try:
    oslo_service_backend.init_backend(
        oslo_service_backend.BackendType.THREADING)
except oslo_service_backend.exceptions.BackendAlreadySelected:
    # NOTE(ralonsoh): this code could be called by other services, like
    # ``oslo-config-generator``, still not migrated.
    warnings.warn('The selected oslo_service backend is "eventlet"')


# pylint: disable=wrong-import-position
import builtins  # noqa: E402
import gettext  # noqa: E402

from neutron._i18n import _ as n_under  # noqa: E402


gettext.install('neutron')


# gettext will install its own translation function, override it to be
# the one from neutron
builtins.__dict__['_'] = n_under



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from neutron.agent.ovn.extensions.bgp import commands
from neutron.services.bgp import constants
from neutron.tests.functional.services import bgp


def _get_unique_name(prefix="test"):
    return f"{prefix}_{uuidutils.generate_uuid()[:8]}"


class SetChassisBgpBridgesCommandTestCase(bgp.BaseBgpSbIdlTestCase):
    def _get_chassis_bgp_bridges(self, chassis_name):
        chassis = self.sb_api.lookup('Chassis_Private', chassis_name)
        bridges_str = chassis.external_ids.get(
            constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY, '')
        if bridges_str:
            return bridges_str.split(',')
        return []

    def test_set_bridges_on_chassis_without_existing_bridges(self):
        chassis = self.add_fake_chassis(_get_unique_name("chassis"), '1.1.1.1')
        bridge_name = _get_unique_name("bridge")

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertFalse(bridges)

        commands.SetChassisBgpBridgesCommand(
            self.sb_api, chassis.name, [bridge_name]).execute(check_error=True)

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([bridge_name], bridges)

    def test_set_bridges_replaces_existing_bridges(self):
        existing_bridge = _get_unique_name("existing_bridge")
        chassis = self.add_fake_chassis(
            _get_unique_name("chassis"), '1.1.1.2',
            external_ids={
                constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY: existing_bridge
            })
        new_bridge = _get_unique_name("new_bridge")

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([existing_bridge], bridges)

        commands.SetChassisBgpBridgesCommand(
            self.sb_api, chassis.name, [new_bridge]).execute(check_error=True)

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([new_bridge], bridges)

    def test_set_bridges_is_idempotent(self):
        bridge_name = _get_unique_name("bridge")
        chassis = self.add_fake_chassis(
            _get_unique_name("chassis"), '1.1.1.3',
            external_ids={
                constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY: bridge_name
            })

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([bridge_name], bridges)

        commands.SetChassisBgpBridgesCommand(
            self.sb_api, chassis.name, [bridge_name]).execute(check_error=True)

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([bridge_name], bridges)

    def test_set_bridges_replaces_multiple_existing(self):
        existing_bridge1 = _get_unique_name("existing_bridge1")
        existing_bridge2 = _get_unique_name("existing_bridge2")
        existing_bridges_str = f"{existing_bridge1},{existing_bridge2}"
        chassis = self.add_fake_chassis(
            _get_unique_name("chassis"), '1.1.1.4',
            external_ids={
                constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY: existing_bridges_str
            })
        new_bridge = _get_unique_name("new_bridge")

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertCountEqual([existing_bridge1, existing_bridge2], bridges)

        commands.SetChassisBgpBridgesCommand(
            self.sb_api, chassis.name, [new_bridge]).execute(check_error=True)

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([new_bridge], bridges)

    def test_set_multiple_bridges_at_once(self):
        chassis = self.add_fake_chassis(_get_unique_name("chassis"), '1.1.1.5')
        bridge1 = _get_unique_name("bridge1")
        bridge2 = _get_unique_name("bridge2")
        bridge3 = _get_unique_name("bridge3")

        commands.SetChassisBgpBridgesCommand(
            self.sb_api, chassis.name,
            [bridge1, bridge2, bridge3]).execute(check_error=True)

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertCountEqual([bridge1, bridge2, bridge3], bridges)

    def test_set_bridges_preserves_other_external_ids(self):
        other_key = 'other-external-id'
        other_value = 'some-value'
        chassis = self.add_fake_chassis(
            _get_unique_name("chassis"), '1.1.1.6',
            external_ids={other_key: other_value})
        bridge_name = _get_unique_name("bridge")

        commands.SetChassisBgpBridgesCommand(
            self.sb_api, chassis.name, [bridge_name]).execute(check_error=True)

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([bridge_name], bridges)

        chassis = self.sb_api.lookup('Chassis_Private', chassis.name)
        self.assertEqual(other_value, chassis.external_ids.get(other_key))

    def test_set_empty_bridges_clears_existing(self):
        existing_bridge = _get_unique_name("existing_bridge")
        chassis = self.add_fake_chassis(
            _get_unique_name("chassis"), '1.1.1.7',
            external_ids={
                constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY: existing_bridge
            })

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertEqual([existing_bridge], bridges)

        commands.SetChassisBgpBridgesCommand(
            self.sb_api, chassis.name, []).execute(check_error=True)

        bridges = self._get_chassis_bgp_bridges(chassis.name)
        self.assertFalse(bridges)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
from ovsdbapp.backend.ovs_idl import event as row_event

from neutron._i18n import _
from neutron.agent.ovn.agent import ovsdb
from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.services.bgp import constants

LOG = log.getLogger(__name__)


def _get_external_ids_list(row, key):
    try:
        value = row.external_ids[key]
    except (KeyError, AttributeError):
        return []

    if not value:
        return []

    return [stripped for item in value.split(',')
            if (stripped := item.strip())]


def _get_bgp_peer_bridges(row):
    return set(_get_external_ids_list(row, constants.AGENT_BGP_PEER_BRIDGES))


def _get_ovn_bridge_mappings(row):
    """Get OVN bridge mappings as {bridge_name: 'network:bridge'} dict."""
    return {
        mapping.split(':')[1]: mapping
        for mapping in _get_external_ids_list(row, 'ovn-bridge-mappings')}


class BGPAgentEvent(row_event.RowEvent):
    """Base class for BGP agent events."""

    def __init__(self, agent_api):
        self.agent_api = agent_api
        super().__init__(self.EVENTS, self.TABLE, None)

    @property
    def bgp_agent(self):
        if not hasattr(self, '_bgp_agent'):
            try:
                self._bgp_agent = self.agent_api[constants.AGENT_BGP_EXT_NAME]
            except KeyError:
                raise RuntimeError(_("BGP agent is not configured"))
        return self._bgp_agent


class LocalOVSEvent(BGPAgentEvent):
    """Base class for local OVS events."""
    TABLE = 'Open_vSwitch'

    def _get_desired_mappings(self, row, old):
        bgp_peer_bridges = set(_get_bgp_peer_bridges(row))
        current_mappings = _get_ovn_bridge_mappings(row)
        old_bgp_peer_bridges = set(_get_bgp_peer_bridges(old))

        bgp_mappings = {bridge: f"{bridge}:{bridge}"
                        for bridge in bgp_peer_bridges}

        # Keep all non-BGP mappings as-is
        non_bgp_mappings = {
            mapping
            for bridge, mapping in current_mappings.items()
            if bridge not in bgp_peer_bridges | old_bgp_peer_bridges
        }

        return (
            bgp_mappings,
            sorted(set(bgp_mappings.values()) | non_bgp_mappings))

    def run(self, event, row, old):
        bgp_mappings, desired_bridge_mappings = (
            self._get_desired_mappings(row, old))
        ovsdb.set_ovn_bridge_mapping(
            self.agent_api.ovs_idl, desired_bridge_mappings)
        if bgp_mappings:
            port_mappings_str = ','.join(bgp_mappings.values())
        else:
            port_mappings_str = ''
        self.agent_api.ovs_idl.db_set(
            'Open_vSwitch', '.',
            external_ids={
                constants.OVN_DYNAMIC_ROUTING_PORT_MAPPING: port_mappings_str}
        ).execute(check_error=True)
        bgp_bridge_names = sorted(_get_bgp_peer_bridges(row))
        self.bgp_agent.set_chassis_bgp_bridges(bgp_bridge_names)


class CreateLocalOVSEvent(LocalOVSEvent):
    EVENTS = (LocalOVSEvent.ROW_CREATE,)

    def match_fn(self, event, row, old):
        if constants.AGENT_BGP_PEER_BRIDGES not in row.external_ids:
            LOG.warning("The BGP bridges are not configured")
            return False
        return True


class UpdateLocalOVSEvent(LocalOVSEvent):
    EVENTS = (LocalOVSEvent.ROW_UPDATE,)

    def match_fn(self, event, row, old):
        bgp_mappings, desired_bridge_mappings = self._get_desired_mappings(
            row, old)
        bm_bridges = sorted(_get_ovn_bridge_mappings(row).values())
        return desired_bridge_mappings != bm_bridges


class NewBgpBridgeEvent(BGPAgentEvent):
    EVENTS = (BGPAgentEvent.ROW_CREATE, BGPAgentEvent.ROW_UPDATE,)
    TABLE = 'Bridge'

    @staticmethod
    def _get_bgp_bridges(idl):
        # The passed object is an OvsdbIdl object and not an API object so
        # we cannot use db_get() or any other API methods here.
        ovs_entries = list(idl.tables['Open_vSwitch'].rows.values())
        bgp_bridges_text = ovs_entries[0].external_ids.get(
            constants.AGENT_BGP_PEER_BRIDGES, '')
        if bgp_bridges_text:
            return bgp_bridges_text.split(',')
        return []

    @staticmethod
    def _is_bgp_bridge(row):
        # We need to use row._idl because we may not have access to the
        # agent_api yet when handling events on startup.
        bgp_bridges = NewBgpBridgeEvent._get_bgp_bridges(row._idl)
        return row.name in bgp_bridges

    @staticmethod
    def _has_nic_iface(row):
        for port in row.ports:
            iface = port.interfaces[0]
            if iface.type in constants.BGP_BRIDGE_NIC_TYPES:
                return True
        return False

    @staticmethod
    def _nic_iface_added(row, old):
        added_ports = set(row.ports) - set(old.ports)
        return any(port.interfaces[0].type in constants.BGP_BRIDGE_NIC_TYPES
                   for port in added_ports)

    def match_fn(self, event, row, old):
        if not super().match_fn(event, row, old):
            return False

        if not self._is_bgp_bridge(row):
            return False

        if event == BGPAgentEvent.ROW_UPDATE:
            if (not hasattr(old, 'ports') or
                    not self._nic_iface_added(row, old)):
                return False

        return self._has_nic_iface(row)

    def run(self, event, row, old):
        bgp_bridge = self.bgp_agent.create_bgp_bridge(row.name)
        self.bgp_agent.watch_port_created_event(bgp_bridge, 'patch')
        # Empty string is for the NIC connecting to the leaf switch
        self.bgp_agent.watch_port_created_event(bgp_bridge, '')
        if bgp_bridge.check_requirements_for_flows_met():
            bgp_bridge.configure_flows()


class PortBindingLrpMacEvent(BGPAgentEvent):
    """Port_Binding update event - set LRP MAC."""
    TABLE = 'Port_Binding'
    EVENTS = (BGPAgentEvent.ROW_CREATE, BGPAgentEvent.ROW_UPDATE,)

    def __init__(self, agent_api):
        super().__init__(agent_api)
        self.chassis = ovsdb.get_own_chassis_name(agent_api.ovs_idl)

    def match_fn(self, event, row, old):
        if not super().match_fn(event, row, old):
            return False
        if row.chassis and row.chassis[0].name != self.chassis:
            return False
        if row.type != ovn_const.PB_TYPE_L3GATEWAY:
            return False
        if constants.LRP_NETWORK_NAME_EXT_ID_KEY not in row.external_ids:
            return False
        try:
            ovn_utils.get_mac_and_ips_from_port_binding(row)
        except ValueError as ve:
            LOG.error("Failed to get MAC address from port binding %s: %s",
                      row, ve)
            return False
        return True

    def run(self, event, row, old):
        network_name = row.external_ids[constants.LRP_NETWORK_NAME_EXT_ID_KEY]
        try:
            bridge = self.bgp_agent.bgp_bridges[network_name]
        except KeyError:
            LOG.warning("No BGP bridge found for network %s", network_name)
            return
        if bridge.check_requirements_for_flows_met():
            bridge.configure_flows()


class BGPBridgePortCreatedEvent(BGPAgentEvent):
    EVENTS = (BGPAgentEvent.ROW_CREATE,)
    TABLE = 'Interface'
    ONETIME = True

    def __init__(self, agent_api, bgp_bridge_name, port_type):
        super().__init__(agent_api)
        self.bgp_bridge_name = bgp_bridge_name
        self.port_type = port_type

    @property
    def key(self):
        return (self.__class__, self.table,
                tuple(self.events), self.bgp_bridge_name)

    def _get_port_bridge(self, port_name):
        # We just need access to BaseOVS
        some_bridge = next(iter(self.bgp_agent.bgp_bridges.values()))
        return some_bridge.ovs_bridge.get_bridge_for_iface(port_name)

    def match_fn(self, event, row, old):
        if not super().match_fn(event, row, old):
            return False

        if row.type != self.port_type:
            return False

        try:
            port_bridge_name = self._get_port_bridge(row.name)
        except StopIteration:
            LOG.warning("No BGP bridge found in agent.")
            return False

        return port_bridge_name == self.bgp_bridge_name

    def run(self, event, row, old):
        port_bridge_name = self._get_port_bridge(row.name)
        bgp_bridge = self.bgp_agent.bgp_bridges[port_bridge_name]
        if bgp_bridge.check_requirements_for_flows_met():
            bgp_bridge.configure_flows()



# Copyright 2015 Cloudbase Solutions.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

from neutron_lib.utils import runtime
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from neutron.agent.linux import utils
from neutron.conf.agent import common as config
from neutron.conf.agent.database import agents_db


LOG = logging.getLogger(__name__)
config.register_root_helper(cfg.CONF)
agents_db.register_db_agents_opts()

INTERFACE_NAMESPACE = 'neutron.interface_drivers'


create_process = utils.create_process
kill_process = utils.kill_process
execute = utils.execute
get_root_helper_child_pid = utils.get_root_helper_child_pid
pid_invoked_with_cmdline = utils.pid_invoked_with_cmdline


def load_interface_driver(conf, get_networks_callback=None):
    """Load interface driver for agents like DHCP or L3 agent.

    :param conf: Driver configuration object
    :param get_networks_callback: A callback to get network information.
                                  This will be passed as additional keyword
                                  argument to the interface driver.
    :raises SystemExit of 1 if driver cannot be loaded
    """

    try:
        loaded_class = runtime.load_class_by_alias_or_classname(
            INTERFACE_NAMESPACE, conf.interface_driver)
        return loaded_class(conf, get_networks_callback=get_networks_callback)
    except ImportError:
        LOG.error("Error loading interface driver '%s'",
                  conf.interface_driver)
        raise SystemExit(1)


def is_agent_down(heart_beat_time):
    return timeutils.is_older_than(heart_beat_time,
                                   cfg.CONF.agent_down_time)


def get_hypervisor_hostname():
    """Get hypervisor hostname

    This logic is implemented following the logic of virGetHostnameImpl
    in libvirt.
    """
    hypervisor_hostname = socket.gethostname()
    if (hypervisor_hostname.startswith('localhost') or
            '.' in hypervisor_hostname):
        return hypervisor_hostname

    try:
        addrinfo = socket.getaddrinfo(host=hypervisor_hostname,
                                      port=None,
                                      family=socket.AF_UNSPEC,
                                      flags=socket.AI_CANONNAME)
        # getaddrinfo returns a list of 5-tuples with;
        #     (family, type, proto, canonname, sockaddr)
        if (addrinfo and addrinfo[0][3] and
                not addrinfo[0][3].startswith('localhost')):
            return addrinfo[0][3]
    except OSError as os_err:
        LOG.warning("Error: %s, occured while querying for fqdn. "
                    "get_hypervisor_hostname will just "
                    "return %s", os_err, hypervisor_hostname)

    return hypervisor_hostname


# TODO(bence romsics): rehome this to neutron_lib.placement.utils
def default_rp_hypervisors(hypervisors, device_mappings,
                           default_hypervisor=None,
                           tunnelled_network_rp_name=None):
    """Fill config option 'resource_provider_hypervisors' with defaults.

    :param hypervisors: Config option 'resource_provider_hypervisors'
        as parsed by oslo.config, that is a dict with keys of physical devices
        and values of hypervisor names.
    :param device_mappings: Device mappings standardized to the list-valued
        format.
    :param default_hypervisor: Default hypervisor hostname. If not set,
        it tries to default to fully qualified domain name (fqdn)
    :param tunnelled_network_rp_name: the resource provider name for tunnelled
        networks; if present, it will be added to the devices list.
    """
    _default_hypervisor = default_hypervisor or get_hypervisor_hostname()
    # device_mappings = {'physnet1': ['br-phy1'], 'physnet2': ['br-phy2'], ...}
    devices = {dev for devs in device_mappings.values() for dev in devs}
    if tunnelled_network_rp_name:
        devices.add(tunnelled_network_rp_name)
    return {device: hypervisors.get(device) or _default_hypervisor
            for device in devices}



# Copyright 2022 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import datetime
import random
import time
from unittest import mock

from oslo_utils import timeutils

from neutron.common.ovn import constants as ovn_const
from neutron.plugins.ml2.drivers.ovn.agent import neutron_agent
from neutron.tests import base
from neutron.tests.unit import fake_resources as fakes


class AgentCacheTestCase(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.agent_cache = neutron_agent.AgentCache(driver=mock.ANY)
        self.addCleanup(self._clean_agent_cache)
        self.agents = {}
        self.num_agents = 10  # Add 10 agents.
        for i in range(self.num_agents):
            agent_type = random.choice(ovn_const.OVN_AGENT_TYPES)
            other_config = {}
            if agent_type == ovn_const.OVN_CONTROLLER_GW_AGENT:
                # 'enable-chassis-as-gw' is mandatory if the controller is
                # a gateway chassis; if not, it will default to
                # 'OVN Controller agent'. Check ``ControllerGatewayAgent``
                # class.
                other_config = {'ovn-cms-options': 'enable-chassis-as-gw'}
            chassis = fakes.FakeOvsdbRow.create_one_ovsdb_row(
                attrs={'other_config': other_config,
                       'hostname': f'host{i:d}',
                       })
            ext_ids = {}
            if agent_type == ovn_const.OVN_METADATA_AGENT:
                ext_ids = {
                    ovn_const.OVN_AGENT_METADATA_ID_KEY: 'chassis' + str(i)}
            elif agent_type == ovn_const.OVN_NEUTRON_AGENT:
                ext_ids = {
                    ovn_const.OVN_AGENT_NEUTRON_ID_KEY: 'chassis' + str(i)}
            chassis_private = fakes.FakeOvsdbRow.create_one_ovsdb_row(
                attrs={'name': 'chassis' + str(i),
                       'other_config': {},
                       'chassis': [chassis],
                       'nb_cfg_timestamp': timeutils.utcnow_ts() * 1000,
                       'external_ids': ext_ids,
                       })
            self.agent_cache.update(agent_type, chassis_private)
            self.agents['chassis' + str(i)] = agent_type

        self.assertEqual(self.num_agents, len(list(self.agent_cache)))
        for agent_class in (neutron_agent.NeutronAgent,
                            neutron_agent.MetadataAgent,
                            neutron_agent.OVNNeutronAgent):
            mock.patch.object(agent_class, 'alive', return_value=True).start()

    def _clean_agent_cache(self):
        del self.agent_cache

    def _list_agents(self):
        self.names_read = []
        for idx, agent in enumerate(self.agent_cache):
            self.names_read.append(agent.agent_id)
            if idx == 5:  # Swap to "_add_and_delete_agents" thread.
                time.sleep(0)

    def _add_and_delete_agents(self):
        self.agent_cache.delete('chassis8')
        chassis = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'other_config': {}})
        chassis_private = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'name': 'chassis10',
                   'chassis': [chassis],
                   'nb_cfg_timestamp': timeutils.utcnow_ts() * 1000})
        self.agent_cache.update(ovn_const.OVN_CONTROLLER_AGENT,
                                chassis_private)

    def test_update_while_iterating_agents(self):
        # TODO(ralonsoh): refactor this test to make it compatible after the
        # eventlet removal.
        self.skipTest('This test is skipped after the eventlet removal and '
                      'needs to be refactored')

        # NOTE(ralonsoh): I'm keeping the old code commented in order to
        # refactor it. It is commented because eventlet library is removed.
        # pool = eventlet.GreenPool(2)
        # pool.spawn(self._list_agents)
        # pool.spawn(self._add_and_delete_agents)
        # pool.waitall()
        # self.assertEqual(list(self.agents.keys()), self.names_read)

    def test_agents_by_chassis_private(self):
        ext_ids = {ovn_const.OVN_AGENT_METADATA_ID_KEY: 'chassis5'}
        chassis_private = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'name': 'chassis5',
                   'external_ids': ext_ids})
        agents = self.agent_cache.agents_by_chassis_private(chassis_private)
        agents = list(agents)
        self.assertEqual(1, len(agents))
        self.assertEqual('chassis5', agents[0].agent_id)

    def test_heartbeat_timestamp_format(self):
        chassis_private = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'name': 'chassis5'})
        agents = self.agent_cache.agents_by_chassis_private(chassis_private)
        agent = list(agents)[0]
        agent.chassis.hostname = 'fake-hostname'
        agent.updated_at = datetime.datetime(
            year=2023, month=2, day=23, hour=1, minute=2, second=3,
            microsecond=456789).replace(tzinfo=datetime.timezone.utc)

        # Verify that both microseconds and timezone are dropped
        self.assertEqual(str(agent.as_dict()['heartbeat_timestamp']),
                         '2023-02-23 01:02:03')

    def test_list_agents_filtering_host_same_type(self):
        for idx in range(len(self.agents)):
            host = f'host{idx:d}'
            agents = self.agent_cache.get_agents(filters={'host': host})
            self.assertEqual(1, len(agents))
            self.assertEqual(host, agents[0].as_dict()['host'])

    def test_list_agents_filtering_host_as_iterable(self):
        hosts = []
        for idx in range(len(self.agents)):
            hosts.append(f'host{idx:d}')

        agents = self.agent_cache.get_agents(filters={'host': hosts})
        self.assertEqual(len(self.agents), len(agents))

    def test_list_agents_filtering_agent_type_same_type(self):
        agent_types = collections.defaultdict(int)
        for _type in self.agents.values():
            agent_types[_type] = agent_types[_type] + 1

        for _type in agent_types:
            agents = self.agent_cache.get_agents(
                filters={'agent_type': _type})
            self.assertEqual(agent_types[_type], len(agents))
            self.assertEqual(_type, agents[0].as_dict()['agent_type'])

    def test_list_agents_filtering_agent_type_as_iterable(self):
        agents = self.agent_cache.get_agents(
            filters={'agent_type': ovn_const.OVN_AGENT_TYPES})
        self.assertEqual(self.num_agents, len(agents))

    @mock.patch.object(neutron_agent, 'LOG')
    def test_list_agents_filtering_wrong_type(self, mock_log):
        agents = self.agent_cache.get_agents(filters={'host': 111})
        self.assertEqual(0, len(agents))
        mock_log.info.assert_called_once()

    def test_list_agents_filtering_same_string_in_filter(self):
        # As reported in LP#2110094, if two registers have the same substring,
        # the filter didn't work.
        # Chassis 1, hostname: compute-0
        chassis = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'other_config': {},
                   'hostname': 'compute-0'})
        chassis_private = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'name': 'chassis1',
                   'other_config': {},
                   'chassis': [chassis],
                   'nb_cfg_timestamp': timeutils.utcnow_ts() * 1000,
                   'external_ids': {}})
        self.agent_cache.update(ovn_const.OVN_CONTROLLER_AGENT,
                                chassis_private)

        # Chassis 2, hostname: dcn1-compute-0
        chassis = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'other_config': {},
                   'hostname': 'dcn1-compute-0'})
        chassis_private = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={'name': 'chassis2',
                   'other_config': {},
                   'chassis': [chassis],
                   'nb_cfg_timestamp': timeutils.utcnow_ts() * 1000,
                   'external_ids': {}})
        self.agent_cache.update(ovn_const.OVN_CONTROLLER_AGENT,
                                chassis_private)

        agents = self.agent_cache.get_agents(
            filters={'host': 'compute-0'})
        self.assertEqual(1, len(agents))

        agents = self.agent_cache.get_agents(
            filters={'host': 'dcn1-compute-0'})
        self.assertEqual(1, len(agents))

        agents = self.agent_cache.get_agents(
            filters={'host': ['compute-0', 'dcn1-compute-0']})
        self.assertEqual(2, len(agents))



#!/usr/bin/env python
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
from wsgiref import simple_server as wsgi_simple_server

from oslo_config import cfg
from oslo_log import log as logging

from neutron.common import config as common_config

LOG = logging.getLogger(__name__)


metadata_opts = [
    cfg.StrOpt('metadata_host', default='127.0.0.1'),
    cfg.IntOpt('metadata_port', default=8775)
]
cfg.CONF.register_opts(metadata_opts)


class FakeMetadata:

    def wsgi_app(self, env, start_response):
        response_headers = [('Content-Type', 'application/json')]
        http_status = '200 OK'
        LOG.info("HTTP_X_INSTANCE_ID: %s", env.get('HTTP_X_INSTANCE_ID'))
        LOG.info("HTTP_X_TENANT_ID: %s", env.get('HTTP_X_TENANT_ID'))
        LOG.info("HTTP_X_INSTANCE_ID_SIGNATURE: %s",
                 env.get('HTTP_X_INSTANCE_ID_SIGNATURE'))

        # Send the headers back for verify the path
        response_headers += [
            ('RESP_INSTANCE_ID', env.get('HTTP_X_INSTANCE_ID')),
            ('RESP_TENANT_ID', env.get('HTTP_X_TENANT_ID')),
            ('RESP_INSTANCE_ID_SIGNATURE',
             env.get('HTTP_X_INSTANCE_ID_SIGNATURE'))]

        start_response(http_status, response_headers)
        return [b"Metadata OK"]


if __name__ == "__main__":
    common_config.register_common_config_options()
    common_config.init(sys.argv[1:])
    common_config.setup_logging()
    metadata_host = cfg.CONF.metadata_host
    metadata_port = cfg.CONF.metadata_port

    LOG.info("Metadata fixture started on port: %s", metadata_port)

    mock_metadata = FakeMetadata()
    wsgi_simple_server.make_server(
        metadata_host, metadata_port, mock_metadata.wsgi_app).serve_forever()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as lib_exc
from neutron_lib.plugins import constants as plugin_constants
from oslo_log import log

from neutron.db import servicetype_db as st_db
from neutron.services.l3_router.service_providers import driver_controller
from neutron.services import provider_configuration


LOG = log.getLogger(__name__)


class DriverController(driver_controller.DriverController):
    """Driver controller for the OVN L3 service plugin.

    This component is responsible for dispatching router requests to L3
    service providers and for performing the bookkeeping about which
    driver is associated with a given router.

    This is not intended to be accessed by the drivers or the l3 plugin.
    All of the methods are marked as private to reflect this.
    """

    def __init__(self, l3_plugin):
        self.l3_plugin = l3_plugin
        self._stm = st_db.ServiceTypeManager.get_instance()
        self._stm.add_provider_configuration(
            plugin_constants.L3, _OvnPlusProviderConfiguration())
        self._load_drivers()

    def _get_provider_for_create(self, context, router):
        """Get provider based on flavor or default provider."""
        if not driver_controller.flavor_specified(router):
            return self.drivers[self.default_provider]
        return self._get_l3_driver_by_flavor(context, router['flavor_id'])


class _OvnPlusProviderConfiguration(
        provider_configuration.ProviderConfiguration):

    def __init__(self):
        # loads up the OVN provider automatically and sets it as default.
        super().__init__(
            svc_type=plugin_constants.L3)
        path = 'neutron.services.ovn_l3.service_providers.ovn.OvnDriver'
        try:
            self.add_provider({'service_type': plugin_constants.L3,
                               'name': 'ovn', 'driver': path, 'default': True})
        except lib_exc.Invalid:
            LOG.debug("Could not add L3 provider ovn, it may have "
                      "already been explicitly defined.")



# Copyright (c) 2017 Fujitsu Limited
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.callbacks import events
from neutron_lib.callbacks import priority_group
from neutron_lib import exceptions
from neutron_lib import fixture
from neutron_lib.services.logapi import constants as log_const

from neutron.services.logapi.common import exceptions as log_exc
from neutron.services.logapi.drivers import base as log_driver_base
from neutron.services.logapi.drivers import manager as driver_mgr
from neutron.tests.unit.services.logapi import base


class TestGetParameter(base.BaseLogTestCase):

    def test__get_param_missing_parameter(self):
        kwargs = {'context': mock.sentinel.context}
        self.assertRaises(log_exc.LogapiDriverException,
                          driver_mgr._get_param,
                          args=[], kwargs=kwargs,
                          name='log_obj', index=1)
        self.assertRaises(log_exc.LogapiDriverException,
                          driver_mgr._get_param,
                          args=[mock.sentinel.context], kwargs={},
                          name='log_obj', index=1)
        self.assertRaises(log_exc.LogapiDriverException,
                          driver_mgr._get_param,
                          args=[], kwargs={'log_obj': mock.sentinel.log_obj},
                          name='context', index=0)


class TestLogDriversManagerBase(base.BaseLogTestCase):

    def setUp(self):
        super().setUp()
        self.config_parse()
        self.setup_coreplugin(load_plugins=False)

    @staticmethod
    def _create_manager_with_drivers(drivers_details):
        for name, driver_details in drivers_details.items():
            class LogDriver(log_driver_base.DriverBase):
                @property
                def is_loaded(self):
                    return driver_details['is_loaded']

            LogDriver(name,
                      driver_details.get('vif_types', []),
                      driver_details.get('vnic_types', []),
                      driver_details.get('supported_logging_types', []))

        return driver_mgr.LoggingServiceDriverManager()


class TestLogDriversManagerMulti(TestLogDriversManagerBase):
    """Test calls happen to all drivers"""
    def test_driver_manager_empty_with_no_drivers(self):
        driver_manager = self._create_manager_with_drivers({})
        self.assertEqual(0, len(driver_manager.drivers))

    def test_driver_manager_empty_with_no_loaded_drivers(self):
        driver_manager = self._create_manager_with_drivers(
            {'driver-A': {'is_loaded': False}})
        self.assertEqual(0, len(driver_manager.drivers))

    def test_driver_manager_with_one_loaded_driver(self):
        driver_manager = self._create_manager_with_drivers(
            {'driver-A': {'is_loaded': True}})
        self.assertEqual(1, len(driver_manager.drivers))

    def test_driver_manager_with_two_loaded_drivers(self):
        driver_manager = self._create_manager_with_drivers(
            {'driver-A': {'is_loaded': True},
             'driver-B': {'is_loaded': True}})
        self.assertEqual(2, len(driver_manager.drivers))


class TestLogDriversManagerLoggingTypes(TestLogDriversManagerBase):
    """Test supported logging types"""
    def test_available_logging_types(self):
        driver_manager = self._create_manager_with_drivers(
            {'driver-A': {'is_loaded': True,
                          'supported_logging_types': ['security_group']},
             'driver-B': {'is_loaded': True,
                          'supported_logging_types':
                              ['security_group', 'firewall']}
             })
        self.assertEqual({'security_group', 'firewall'},
                         driver_manager.supported_logging_types)


class TestLogDriversCalls(TestLogDriversManagerBase):
    """Test log driver calls"""

    def setUp(self):
        super().setUp()
        self.driver_manager = self._create_manager_with_drivers(
            {'driver-A': {'is_loaded': True}})

    def test_implemented_call_methods(self):
        for method in log_const.LOG_CALL_METHODS:
            with mock.patch.object(log_driver_base.DriverBase, method) as \
                    method_fnc:
                context = mock.sentinel.context
                log_obj = mock.sentinel.log_obj
                self.driver_manager.call(
                    method, context=context, log_objs=[log_obj])
                method_fnc.assert_called_once_with(
                    context=context, log_objs=[log_obj])

    def test_not_implemented_call_methods(self):
        context = mock.sentinel.context
        log_obj = mock.sentinel.log_obj
        self.assertRaises(exceptions.DriverCallError, self.driver_manager.call,
                          'wrong_method', context=context, log_objs=[log_obj])


class TestHandleResourceCallback(TestLogDriversManagerBase):
    """Test handle resource callback"""

    def setUp(self):
        super().setUp()
        self._cb_mgr = mock.Mock()
        self.useFixture(fixture.CallbackRegistryFixture(
            callback_manager=self._cb_mgr))
        self.driver_manager = driver_mgr.LoggingServiceDriverManager()

    def test_subscribe_resources_cb(self):

        class FakeResourceCB1(driver_mgr.ResourceCallBackBase):
            def handle_event(self, resource, event, trigger, **kwargs):
                pass

        class FakeResourceCB2(driver_mgr.ResourceCallBackBase):
            def handle_event(self, resource, event, trigger, **kwargs):
                pass

        driver_mgr.RESOURCE_CB_CLASS_MAP = {'fake_resource1': FakeResourceCB1,
                                            'fake_resource2': FakeResourceCB2}

        self.driver_manager._setup_resources_cb_handle()

        fake_resource_cb1 = FakeResourceCB1(
            'fake_resource1', self.driver_manager.call)
        fake_resource_cb2 = FakeResourceCB2(
            'fake_resource2', self.driver_manager.call)
        assert_calls = [
            mock.call(
                fake_resource_cb1.handle_event,
                'fake_resource1', events.AFTER_CREATE,
                priority_group.PRIORITY_DEFAULT, False),
            mock.call(
                fake_resource_cb1.handle_event,
                'fake_resource1', events.AFTER_UPDATE,
                priority_group.PRIORITY_DEFAULT, False),
            mock.call(
                fake_resource_cb1.handle_event,
                'fake_resource1', events.AFTER_DELETE,
                priority_group.PRIORITY_DEFAULT, False),
            mock.call(
                fake_resource_cb2.handle_event,
                'fake_resource2', events.AFTER_CREATE,
                priority_group.PRIORITY_DEFAULT, False),
            mock.call(
                fake_resource_cb2.handle_event,
                'fake_resource2', events.AFTER_UPDATE,
                priority_group.PRIORITY_DEFAULT, False),
            mock.call(
                fake_resource_cb2.handle_event,
                'fake_resource2', events.AFTER_DELETE,
                priority_group.PRIORITY_DEFAULT, False),
        ]
        self._cb_mgr.subscribe.assert_has_calls(assert_calls)



# Copyright 2013 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.objects.port.extensions import port_security
from neutron.tests.unit.objects import test_base as obj_test_base
from neutron.tests.unit import testlib_api


class PortSecurityIfaceObjTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = port_security.PortSecurity


class PortSecurityDbObjTestCase(obj_test_base.BaseDbObjectTestCase,
                                testlib_api.SqlTestCase):

    _test_class = port_security.PortSecurity

    def setUp(self):
        super().setUp()
        network_id = self._create_test_network_id()
        for obj in self.db_objs:
            self._create_test_port(
                id=obj['port_id'], network_id=network_id)



# Copyright (c) 2024 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import qinq as apidef
from neutron_lib.api import extensions as api_extensions
from neutron_lib.api import validators
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def _disable_extension_by_config(aliases):
    if cfg.CONF.vlan_qinq is False:
        if apidef.ALIAS in aliases:
            aliases.remove(apidef.ALIAS)
        LOG.info('Disabled VLAN QinQ extension.')


def get_qinq(network):
    """Get the value of vlan_qinq from a network if set.

    :param network: The network dict to retrieve the value of vlan_qinq
        from.
    :returns: The value of vlan_qinq from the network dict if set in
        the dict, otherwise False is returned.
    """
    return (network[apidef.QINQ_FIELD]
            if (apidef.QINQ_FIELD in network and
                validators.is_attr_set(network[apidef.QINQ_FIELD]))
            else False)


class Qinq(api_extensions.APIExtensionDescriptor):
    """Extension class supporting vlan QinQ networks."""

    api_definition = apidef



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from importlib import util as imp_util
import os

from neutron.db import migration as db_migration
from neutron.db.migration.alembic_migrations.versions import kilo_initial
from neutron.db.migration import cli as migration_cli
from neutron.tests import base


class TestDBMigration(base.BaseTestCase):

    def test_current_release_is_in_releases(self):
        self.assertIn(migration_cli.CURRENT_RELEASE,
                      migration_cli.RELEASES)

    def test_current_release_not_in_neutron_milestones(self):
        self.assertNotIn(migration_cli.CURRENT_RELEASE,
                         db_migration.NEUTRON_MILESTONES)

    def test_neutron_milestones_present_in_releases(self):
        # The RELEASES tuple must have all NEUTRON_MILESTONES values plus
        # the CURRENT_RELEASE one.
        milestones = set(db_migration.NEUTRON_MILESTONES)
        releases = set(migration_cli.RELEASES)
        self.assertEqual({migration_cli.CURRENT_RELEASE},
                         releases - milestones)

    def test_check_all_milestones_are_tagged(self):
        neutron_milestones = set(db_migration.NEUTRON_MILESTONES)
        initial_migration_file = os.path.abspath(kilo_initial.__file__)
        base_dir_migrations = os.path.dirname(initial_migration_file)
        for subdir, dirs, mod_files in os.walk(base_dir_migrations):
            # Read all files present in
            # ``./neutron/db/migration/alembic_migrations``
            for mod_file in mod_files:
                mod_name, file_ext = os.path.splitext(
                    os.path.split(mod_file)[-1])
                mod_path = os.path.join(subdir, mod_file)
                last_dir = os.path.basename(subdir)
                # This test only checks the expand migrations.
                if (file_ext.lower() == '.py' and
                        not mod_name.startswith('_') and
                        last_dir == 'expand'):
                    spec = imp_util.spec_from_file_location(mod_name, mod_path)
                    mod = imp_util.module_from_spec(spec)
                    spec.loader.exec_module(mod)
                    neutron_milestone = getattr(mod, 'neutron_milestone', None)
                    if neutron_milestone:
                        # If the file has the ``neutron_milestone`` variable,
                        # check that is present in the ``NEUTRON_MILESTONES``
                        # list.
                        neutron_milestone = neutron_milestone.pop()
                        try:
                            neutron_milestones.remove(neutron_milestone)
                        except KeyError:
                            self.fail(
                                'Milestone %s not present in the milestones '
                                'list; this milestone has been defined more '
                                'than once')

        # Once all expand migration scripts have been read, all migration
        # milestones must be removed from the ``NEUTRON_MILESTONES`` list.
        self.assertEqual(set(), neutron_milestones)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from importlib.metadata import entry_points

from oslo_config import cfg

from neutron._i18n import _


SYNC_ENTRYPOINTS = 'neutron.ovn.db_sync'
MIGRATION_ENTRYPOINTS = 'neutron.ovn.db_migration'

sync_entrypoints = {
    entrypoint.name: entrypoint
    for entrypoint in entry_points(group=SYNC_ENTRYPOINTS)
}
migration_entrypoints = {
    entrypoint.name: entrypoint
    for entrypoint in entry_points(group=MIGRATION_ENTRYPOINTS)
}

INSTALLED_SYNC_PLUGINS = list(sync_entrypoints)
INSTALLED_MIGRATION_PLUGINS = list(migration_entrypoints)


CORE_OPTS = [
    cfg.StrOpt('sync_plugin',
               choices=INSTALLED_SYNC_PLUGINS,
               help=(_("The subproject to execute the command against. "
                       "Can be one of: '%s'.")
                     % "', '".join(INSTALLED_SYNC_PLUGINS))),
    cfg.StrOpt('migration_plugin',
               choices=INSTALLED_MIGRATION_PLUGINS,
               help=(_("The neutron db migration plugin to execute. "
                       "Can be one of: '%s'.")
                     % "', '".join(INSTALLED_MIGRATION_PLUGINS))),
]


def register_ovn_db_sync_cli_opts(conf):
    conf.register_cli_opts(CORE_OPTS)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import vlog

from neutron.common import utils as common_utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.services.bgp import ovn as bgp_ovn
from neutron.tests.functional import base


class OvnNbIdlWithUniqueConnection(bgp_ovn.BgpOvnNbIdl):
    # because the ovsdb connection is a class attribute, we cannot have
    # two connections in the same process
    @property
    def ovsdb_connection(self):
        return self._ovsdb_connection

    @ovsdb_connection.setter
    def ovsdb_connection(self, value):
        self._ovsdb_connection = value


class OvnNbIdlTest(bgp_ovn.OvnNbIdl):
    api_cls = OvnNbIdlWithUniqueConnection


class TestOvnNbIdl(base.TestOVNFunctionalBase):
    """Test OvnNbIdl read-write operations."""

    def setUp(self):
        super().setUp()
        self.nb_connection = ovn_conf.get_ovn_nb_connection()
        self.nb_idl = bgp_ovn.OvnNbIdl(self.nb_connection)
        self.addCleanup(self._cleanup)
        self.nb_bgp_api = self.nb_idl.start(timeout=10)

    def _cleanup(self):
        if hasattr(self, 'nb_bgp_api') and self.nb_bgp_api:
            try:
                self.nb_bgp_api.ovsdb_connection.stop(timeout=5)
                self.nb_bgp_api.__class__._ovsdb_connection = None
            except Exception:
                pass

    def test_read_write_operations(self):
        """Test NB IDL read and write operations."""
        # Test write: create logical switch
        ls_name = f"test_ls_{uuidutils.generate_uuid()}"
        ls = self.nb_bgp_api.ls_add(ls_name).execute(check_error=True)
        self.assertEqual(ls.name, ls_name)

        # Test read: list logical switches
        ls_list = self.nb_bgp_api.ls_list().execute(check_error=True)
        ls_names = [switch.name for switch in ls_list]
        self.assertIn(ls_name, ls_names)

    def test_leader_only_is_true(self):
        self.assertTrue(self.nb_idl.leader_only)


class TestOvnSbIdl(base.TestOVNFunctionalBase):
    """Test OvnSbIdl read-write operations."""

    def setUp(self):
        super().setUp()
        self.sb_connection = ovn_conf.get_ovn_sb_connection()
        # Register Encap table too so we can create a chassis
        self.addCleanup(self._cleanup)
        bgp_ovn.OvnSbIdl.tables = ('Chassis', 'Encap')
        self.sb_idl = bgp_ovn.OvnSbIdl(self.sb_connection)
        self.sb_bgp_api = self.sb_idl.start(timeout=10)

    def _cleanup(self):
        if hasattr(self, 'sb_bgp_api') and self.sb_bgp_api:
            try:
                self.sb_bgp_api.ovsdb_connection.stop(timeout=5)
                self.sb_bgp_api.__class__._ovsdb_connection = None
            except Exception:
                pass
        bgp_ovn.OvnSbIdl.tables = bgp_ovn.OVN_SB_TABLES

    def test_read_write_operations(self):
        """Test SB IDL read and write operations."""
        # Test write: add chassis
        chassis_name = f"test_chassis_{uuidutils.generate_uuid()}"
        hostname = f"{chassis_name}.example.com"
        self.sb_bgp_api.chassis_add(
            chassis_name, ['geneve'], '192.168.1.100',
            hostname=hostname).execute(check_error=True)

        # Test read: list chassis
        chassis_list = self.sb_bgp_api.chassis_list().execute(check_error=True)
        self.assertIn(chassis_name, [chassis.name for chassis in chassis_list])
        self.assertEqual(len(chassis_list), 1)

    def test_leader_only_is_false(self):
        self.assertFalse(self.sb_idl.leader_only)


class TestBgpOvnLocking(base.TestOVNFunctionalBase):
    """Test BGP OVN locking mechanism."""

    def setUp(self):
        super().setUp()
        self._enable_jsonrpc_debug_logging()
        self.nb_connection = ovn_conf.get_ovn_nb_connection()
        self.addCleanup(self._cleanup)

        # Create two IDL instances to test locking
        self.nb_idl1 = OvnNbIdlTest(self.nb_connection)
        self.nb_bgp_api1 = self.nb_idl1.start(timeout=10)

        self.nb_idl2 = OvnNbIdlTest(self.nb_connection)
        self.nb_bgp_api2 = self.nb_idl2.start(timeout=10)

    def _enable_jsonrpc_debug_logging(self):
        vlog.use_python_logger(max_level=vlog.DEBUG)
        vlog.vlog.Vlog.set_level('jsonrpc', 'any', 'dbg')

        def _restore():
            vlog.vlog.Vlog.set_level('jsonrpc', 'any', 'info')
            vlog.reset_logger()

        self.addCleanup(_restore)

    def _cleanup(self):
        for api in [getattr(self, 'nb_bgp_api1', None),
                    getattr(self, 'nb_bgp_api2', None)]:
            if api:
                try:
                    api.ovsdb_connection.stop(timeout=5)
                except Exception:
                    pass

    def test_locking_mechanism(self):
        """Test BGP topology locking mechanism.

        The NB API uses locking by default so one API should get the lock
        once connected.
        """
        def wait_for_lock():
            return self.nb_bgp_api1.has_lock != self.nb_bgp_api2.has_lock

        # Wait for some API to get the lock
        common_utils.wait_until_true(
            wait_for_lock,
            timeout=5,
            exception=AssertionError("Lock was not obtained by either API")
        )

        api_with_lock = None
        # Make sure the other API does not have the lock
        if self.nb_bgp_api1.has_lock:
            self.assertFalse(self.nb_bgp_api2.has_lock)
            api_with_lock = self.nb_bgp_api1
            api_without_lock = self.nb_bgp_api2

        if self.nb_bgp_api2.has_lock:
            self.assertFalse(self.nb_bgp_api1.has_lock)
            api_with_lock = self.nb_bgp_api2
            api_without_lock = self.nb_bgp_api1

        self.assertIsNotNone(api_with_lock)

        # Disconnect first API and check that second API can acquire the lock
        api_with_lock.ovsdb_connection.stop(timeout=5)

        common_utils.wait_until_true(
            lambda: api_without_lock.has_lock,
            timeout=5,
            exception=AssertionError("Second API did not acquire lock")
        )



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from neutron.agent.ovn.extensions.bgp import bridge
from neutron.agent.ovsdb import impl_idl
from neutron.common import utils
from neutron.services.bgp import constants
from neutron.services.bgp import ovn as bgp_ovn
from neutron.tests.common import net_helpers
from neutron.tests.functional.agent.ovn.extensions import bgp as test_bgp
from neutron.tests.functional.services import bgp as base


class FakeAgentApi:
    def __init__(self, sb_idl):
        self.sb_idl = sb_idl
        self.ovs_idl = impl_idl.api_factory()


class FakeBgpAgentApi:
    def __init__(self, sb_idl):
        self.agent_api = FakeAgentApi(sb_idl)


class BgpTestCaseWithIdls(base.BaseBgpIDLTestCase):
    schemas = ['OVN_Northbound', 'OVN_Southbound']

    def setUp(self):
        bgp_ovn.OvnSbIdl.tables = (
            'Port_Binding', 'Chassis', 'Chassis_Private', 'Encap')
        try:
            super().setUp()
        finally:
            bgp_ovn.OvnSbIdl.tables = bgp_ovn.OVN_SB_TABLES


class BGPChassisBridgeTestCase(BgpTestCaseWithIdls):
    def setUp(self):
        super().setUp()
        self.ovs_api = impl_idl.api_factory()
        self.test_bridge = self.useFixture(
            net_helpers.OVSBridgeFixture()).bridge

        patch_bridge = self.useFixture(
            net_helpers.OVSBridgeFixture()).bridge

        self.fake_nic = self.useFixture(net_helpers.VethFixture()).ports[0]

        bgp_patch_port_name = utils.get_rand_name(max_length=14,
                                                  prefix='bgp-pp')
        peer_patch_port_name = utils.get_rand_name(max_length=14,
                                                   prefix='peer-pp')
        with self.ovs_api.transaction(check_error=True) as txn:
            txn.add(self.ovs_api.add_port(
                self.test_bridge.br_name,
                bgp_patch_port_name, type='patch',
                options={'peer': peer_patch_port_name}))
            txn.add(self.ovs_api.add_port(
                patch_bridge.br_name,
                peer_patch_port_name, type='patch',
                options={'peer': bgp_patch_port_name}))
            txn.add(self.ovs_api.add_port(
                self.test_bridge.br_name, self.fake_nic.name))

        self.bgp_bridge = bridge.BGPChassisBridge(
            FakeBgpAgentApi(self.sb_api),
            self.test_bridge.br_name)

        # Wait for OVS to assign the patch port an ofport
        utils.wait_until_true(
            lambda: (self.bgp_bridge.patch_port_ofport
                     is not None),
            sleep=0.1,
            timeout=5,
            exception=Exception("Patch port ofport not found"))

    @staticmethod
    def _dump_flows(bridge):
        flows_str = bridge.ovs_bridge.dump_flows_for()
        return flows_str.splitlines() if flows_str else []

    def test_patch_port_ofport(self):
        ofport = self.bgp_bridge.patch_port_ofport
        self.assertIn(ofport, [1, 2])

    def test_configure_flows_complete_has_all_expected_flows(self):
        self.bgp_bridge.bgp_agent_api.hostdev_ips = [
            netaddr.IPNetwork('172.16.1.1/30'),
            netaddr.IPNetwork('2001:db8::1/64'),
            netaddr.IPNetwork('192.168.1.10/32'),
            netaddr.IPNetwork('10.0.0.1/32'),
            netaddr.IPNetwork('2001:db8:eee::1/128'),
        ]

        chassis_name = 'chassis-test'
        chassis = self.sb_api.chassis_add(
            chassis_name,
            ['geneve'],
            '172.24.4.10',
            hostname=chassis_name,
        ).execute(check_error=True)
        self.sb_api.db_create(
            'Chassis_Private', name=chassis_name,
            chassis=chassis.uuid,
        ).execute(check_error=True)

        port_name = 'lrp-test'

        pb_wait_event = test_bgp.WaitForPortBindingEvent(port_name)
        self.sb_api.idl.notify_handler.watch_event(pb_wait_event)

        lrp_ext_ids = {
            constants.LRP_NETWORK_NAME_EXT_ID_KEY: self.test_bridge.br_name}
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.lr_add(router='lr-test',
                                       options={'chassis': chassis.name}))
            txn.add(self.nb_api.lrp_add(router='lr-test',
                                        port=port_name,
                                        mac='aa:bb:cc:dd:ee:ff',
                                        networks=['192.168.1.10/32'],
                                        external_ids=lrp_ext_ids))

        self.assertTrue(pb_wait_event.wait())

        self.bgp_bridge.configure_flows()

        flows = self._dump_flows(self.bgp_bridge)
        flow_strings = ' '.join(flows)

        # 1. ARP and ICMPv6 flows
        self.assertIn('arp actions=NORMAL', flow_strings)
        self.assertIn(f'icmp6,in_port={self.bgp_bridge.nic_ofport},'
                      'icmp_type=133 actions=NORMAL,'
                      'mod_dl_dst:aa:bb:cc:dd:ee:ff,'
                      f'output:{self.bgp_bridge.patch_port_ofport}',
                      flow_strings)
        self.assertIn(f'icmp6,in_port={self.bgp_bridge.nic_ofport},'
                      'icmp_type=134 actions=NORMAL,'
                      'mod_dl_dst:aa:bb:cc:dd:ee:ff,'
                      f'output:{self.bgp_bridge.patch_port_ofport}',
                      flow_strings)
        self.assertIn(f'icmp6,in_port={self.bgp_bridge.nic_ofport},'
                      'icmp_type=135 actions=NORMAL,'
                      'mod_dl_dst:aa:bb:cc:dd:ee:ff,'
                      f'output:{self.bgp_bridge.patch_port_ofport}',
                        flow_strings)
        self.assertIn(f'icmp6,in_port={self.bgp_bridge.nic_ofport},'
                      'icmp_type=136 actions=NORMAL,'
                      'mod_dl_dst:aa:bb:cc:dd:ee:ff,'
                      f'output:{self.bgp_bridge.patch_port_ofport}',
                        flow_strings)

        # 2. Host IP flows (IPv4)
        self.assertIn('nw_dst=192.168.1.10 actions=NORMAL', flow_strings)
        self.assertIn('nw_dst=10.0.0.1 actions=NORMAL', flow_strings)

        # 3. Host IP flows (IPv6)
        self.assertIn('ipv6_dst=2001:db8:eee::1 actions=NORMAL', flow_strings)
        self.assertIn('ipv6_dst=2001:db8::1 actions=NORMAL', flow_strings)

        # 4. IPv6 link-local traffic
        self.assertIn('ipv6_dst=fe80::/64 actions=NORMAL', flow_strings)

        # 5. Default flow
        self.assertIn('priority=0 actions=NORMAL', flow_strings)

        # 6. LRP MAC rewrite flow
        self.assertIn(
            f"in_port={self.bgp_bridge.nic_ofport} actions="
            f"mod_dl_dst:aa:bb:cc:dd:ee:ff,"
            f"output:{self.bgp_bridge.patch_port_ofport}",
            flow_strings)



# Copyright (c) 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import signal
import tempfile

from oslo_config import cfg

from neutron.agent.common import async_process
from neutron.agent.linux import external_process as ep
from neutron.agent.linux import utils as agent_utils
from neutron.common import utils as common_utils
from neutron.tests.common import net_helpers
from neutron.tests.functional import base as functional_base


class ProcessManagerTestCase(functional_base.BaseSudoTestCase):

    def _create_sleep_process(self, time=None):
        if time is None:
            cmd = ['sleep', 'infinity']
        else:
            cmd = ['sleep', str(time)]
        process = async_process.AsyncProcess(cmd)
        process.start()

        with tempfile.NamedTemporaryFile('w+', delete=False) as pid_file:
            pid_file.write(process.pid)
        os.chmod(pid_file.name, 0o777)
        uuid = 'sleep infinity'
        namespace = self.useFixture(net_helpers.NamespaceFixture()).name
        return ep.ProcessManager(cfg.CONF, uuid, namespace,
                                 pid_file=pid_file.name)

    def test__kill_process(self):
        pm = self._create_sleep_process()
        self.assertTrue(pm.active)

        pm._kill_process(pm.get_kill_cmd(int(signal.SIGKILL), pm.pid), pm.pid)
        # Delete the PID file used by ``pm.active``.
        agent_utils.delete_if_exists(pm.get_pid_file_name())
        self.assertFalse(pm.active)

    def test__kill_process_process_not_present(self):
        pm = self._create_sleep_process(time=0)
        # "sleep 0" should end immediately, but we add an active wait of 3
        # seconds just to avoid any race condition.
        try:
            common_utils.wait_until_true(lambda: not pm.active, timeout=3)
        except common_utils.WaitTimeout:
            self.fail('The process "sleep 0" (PID: %s) did not finish' %
                      pm.pid)

        # '_kill_process' should not raise any exception.
        pm._kill_process(pm.get_kill_cmd(int(signal.SIGKILL), pm.pid), pm.pid)
        self.assertFalse(pm.active)



# Copyright 2025 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as neutron_context
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_utils import strutils
from ovsdbapp.backend.ovs_idl import event as row_event

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils


class LogicalRouterPortEvent(row_event.RowEvent):
    """Logical_Router_Port create/delete event.

    If a Logical_Router_Port is deleted or added, first check if this LRP is a
    gateway port or not. Then update the corresponding network (or networks)
    HA_Chassis_Group, matching the Logical_Router Gateway_Chassis.
    See LP#2125553.
    """
    def __init__(self, driver):
        self.driver = driver
        self.l3_plugin = directory.get_plugin(constants.L3)
        self.admin_context = neutron_context.get_admin_context()
        table = 'Logical_Router_Port'
        events = (self.ROW_CREATE, self.ROW_DELETE)
        super().__init__(events, table, None)

    def match_fn(self, event, row, old):
        try:
            ls_name = row.external_ids[ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY]
            lr_name = row.external_ids[ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY]
        except KeyError:
            # The Logical_Router_Port doesn't have the Neutron owned resources
            # external_ids ("neutron:network_name", "neutron:router_name").
            return False

        if event == self.ROW_DELETE:
            # Check if the LR has another port in the same network. If that is
            # the case, do nothing.
            lr = self.driver._nb_ovn.lookup('Logical_Router', lr_name,
                                            default=None)
            if not lr:
                # The LRP has been deleted along with the LR. That happens
                # only with the GW LRPs. No action needed.
                return False

            for lrp in (lrp for lrp in lr.ports if lrp.name != row.name):
                if (ls_name == lrp.external_ids[
                        ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY]):
                    return False
            return True

        # event == self.ROW_CREATE
        return True

    def run(self, event, row, old=None):
        ext_gw = row.external_ids.get(ovn_const.OVN_ROUTER_IS_EXT_GW)
        router_id = utils.get_neutron_name(
            row.external_ids[ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY])
        net_id = utils.get_neutron_name(
            row.external_ids[ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY])
        if event == self.ROW_DELETE:
            if not strutils.bool_from_string(ext_gw):  # LRP internal port.
                self.l3_plugin._ovn_client.unlink_network_ha_chassis_group(
                    net_id)
            else:  # LRP gateway port.
                self.l3_plugin._ovn_client.update_router_ha_chassis_group(
                    self.admin_context, router_id)

        else:  # event == self.ROW_CREATE
            if not strutils.bool_from_string(ext_gw):  # LRP internal port.
                self.l3_plugin._ovn_client.link_network_ha_chassis_group(
                    self.admin_context, net_id, router_id)
            else:  # LRP gateway port.
                self.l3_plugin._ovn_client.update_router_ha_chassis_group(
                    self.admin_context, router_id)


class LogicalRouterPortGatewayChassisEvent(row_event.RowEvent):
    """Logical_Router_Port Gateway_Chassis change event.

    When the Gateway_Chassis list of a Logical_Router_Port changes, it is
    needed to update the linked HA_Chassis_Group registers.
    """
    def __init__(self, driver):
        self.driver = driver
        self.l3_plugin = directory.get_plugin(constants.L3)
        self.admin_context = neutron_context.get_admin_context()
        table = 'Logical_Router_Port'
        events = (self.ROW_UPDATE, )
        super().__init__(events, table, None)

    def match_fn(self, event, row, old):
        if hasattr(old, 'gateway_chassis'):
            # NOTE: when a Gateway_Chassis register is deleted, is no longer
            # present in the old.gateway_chassis list.
            return True

        return False

    def run(self, event, row, old=None):
        lr_name = row.external_ids.get(ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY)
        router_id = utils.get_neutron_name(lr_name)
        self.l3_plugin._ovn_client.update_router_ha_chassis_group(
            self.admin_context, router_id)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.db import api as db_api
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.db import ovn_l3_hamode_db as ovn_l3_ha
from neutron.db import ovn_revision_numbers_db as db_rev
from neutron.extensions import revisions
from neutron.objects import router as l3_obj
from neutron.services.l3_router.service_providers import base
from neutron.services.portforwarding import constants as pf_consts


LOG = logging.getLogger(__name__)


@registry.has_registry_receivers
class OvnDriver(base.L3ServiceProvider,
                ovn_l3_ha.OVN_L3_HA_db_mixin):
    ha_support = base.MANDATORY
    distributed_support = base.MANDATORY

    @registry.receives(resources.ROUTER, [events.PRECOMMIT_CREATE])
    def _process_router_create_precommit(self, resource, event, trigger,
                                         payload):
        context = payload.context
        context.session.flush()
        router_id = payload.resource_id
        router_db = payload.metadata['router_db']
        router = payload.states[0]
        if not utils.is_ovn_provider_router(router):
            return

        # NOTE(ralonsoh): the "distributed" flag is a static configuration
        # parameter that needs to be defined only during the router creation.
        extra_attr = router_db['extra_attributes']
        extra_attr.distributed = ovn_conf.is_ovn_distributed_floating_ip()

        db_rev.create_initial_revision(
            context, router_id, ovn_const.TYPE_ROUTERS,
            std_attr_id=router_db.standard_attr_id)

    @registry.receives(resources.ROUTER, [events.AFTER_CREATE])
    def _process_router_create(self, resource, event, trigger, payload):
        router = payload.states[0]
        if not utils.is_ovn_provider_router(router):
            return
        context = payload.context
        try:
            self.l3plugin._ovn_client.create_router(context, router)
        except Exception:
            with excutils.save_and_reraise_exception():
                # Delete the logical router
                LOG.exception('Unable to create lrouter %s', router['id'])
                self.l3plugin.delete_router(context, router['id'])

    @registry.receives(resources.ROUTER, [events.AFTER_UPDATE])
    def _process_router_update(self, resource, event, trigger, payload):
        router_id = payload.resource_id
        original = payload.states[0]
        updated = payload.states[1]
        if not utils.is_ovn_provider_router(original):
            # flavor_id attribute is not allowed in router PUTs, so we only
            # need to check the original router
            return
        context = payload.context
        try:
            self.l3plugin._ovn_client.update_router(context, updated, original)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception('Unable to update lrouter %s', router_id)
                revert_router = {'router': original}
                self.l3plugin.update_router(context, router_id, revert_router)

    @registry.receives(resources.ROUTER, [events.AFTER_DELETE])
    def _process_router_delete(self, resource, event, trigger, payload):
        router_id = payload.resource_id
        router = payload.states[0]
        if not utils.is_ovn_provider_router(router):
            return
        context = payload.context
        try:
            self.l3plugin._ovn_client.delete_router(context, router_id)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception('Unable to delete lrouter %s', router['id'])
                self.l3plugin.create_router(context, {'router': router})

    @registry.receives(resources.ROUTER_INTERFACE, [events.AFTER_CREATE])
    def _process_add_router_interface(self, resource, event, trigger, payload):
        router = payload.states[0]
        if not utils.is_ovn_provider_router(router):
            return
        context = payload.context
        port = payload.metadata['port']
        subnets = payload.metadata['subnets']
        router_interface_info = self.l3plugin._make_router_interface_info(
            router.id, port['project_id'], port['id'], port['network_id'],
            subnets[-1]['id'], [subnet['id'] for subnet in subnets])
        try:
            self.l3plugin._ovn_client.create_router_port(context, router.id,
                                                         router_interface_info)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception('Unable to add router interface to lrouter %s. '
                              'Interface info: %s', router['id'],
                              router_interface_info)
                self.l3plugin.remove_router_interface(context, router.id,
                                                      router_interface_info)

    @registry.receives(resources.ROUTER_INTERFACE, [events.AFTER_DELETE])
    def _process_remove_router_interface(self, resource, event, trigger,
                                         payload):
        router = payload.states[0]
        if not utils.is_ovn_provider_router(router):
            return
        context = payload.context
        port = payload.metadata['port']
        subnet_ids = payload.metadata['subnet_ids']
        router_interface_info = self.l3plugin._make_router_interface_info(
            router.id, port['project_id'], port['id'], port['network_id'],
            subnet_ids[0], subnet_ids)
        try:
            self.l3plugin._ovn_client.delete_router_port(context, port['id'],
                                                         subnet_ids=subnet_ids)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception('Unable to remove router interface from lrouter '
                              '%s. Interface info: %s', router['id'],
                              router_interface_info)
                self.l3plugin.add_router_interface(
                    context, router.id, payload.metadata['interface_info'])

    def _create_floatingip_initial_revision(self, context, floatingip_db):
        if not floatingip_db.router_id:
            return
        # We get the router with elevated context because floating IPs may be
        # created to be associated with a router created by a different
        # project. Please see
        # https://review.opendev.org/c/openstack/neutron/+/2727B09
        router = self.l3plugin.get_router(context.elevated(),
                                          floatingip_db.router_id)
        if not utils.is_ovn_provider_router(router):
            return
        db_rev.create_initial_revision(
            context, floatingip_db.id, ovn_const.TYPE_FLOATINGIPS,
            may_exist=True, std_attr_id=floatingip_db.standard_attr_id)

    @registry.receives(resources.FLOATING_IP,
                       [events.PRECOMMIT_CREATE, events.PRECOMMIT_UPDATE])
    def _process_floatingip_create_update_precommit(self, resource, event,
                                                    trigger, payload):
        context = payload.context
        floatingip_db = payload.desired_state
        self._create_floatingip_initial_revision(context, floatingip_db)

    @registry.receives(pf_consts.PORT_FORWARDING, [events.AFTER_CREATE])
    def _process_portforwarding_create(self, resource, event, trigger,
                                       payload):
        context = payload.context
        pf_obj = payload.states[0]
        with db_api.CONTEXT_WRITER.using(context):
            fip_db = l3_obj.FloatingIP.get_object(
                context, id=pf_obj.floatingip_id).db_obj
            self._create_floatingip_initial_revision(context, fip_db)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_CREATE])
    def _process_floatingip_create(self, resource, event, trigger, payload):
        revision_row = db_rev.get_revision_row(payload.context,
                                               payload.resource_id)
        if not revision_row:
            return
        # The floating ip dictionary that is sent by the L3 DB plugin in the
        # notification doesn't include the revision number yet. We add it here
        # to the dictionary that is passed to the ovn client
        floatingip = copy.deepcopy(payload.states[0])
        floatingip[revisions.REVISION] = revision_row.revision_number
        qos_policy_id = payload.request_body['floatingip'].get('qos_policy_id')
        if qos_policy_id and 'qos_policy_id' not in floatingip:
            floatingip['qos_policy_id'] = qos_policy_id
        self.l3plugin._ovn_client.create_floatingip(payload.context,
                                                    floatingip)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_UPDATE])
    def _process_floatingip_update(self, resource, event, trigger, payload):
        if not db_rev.get_revision_row(payload.context, payload.resource_id):
            return
        fip = payload.states[1]
        old_fip = payload.states[0]
        fip_request = payload.request_body
        if fip_request:
            self.l3plugin._ovn_client.update_floatingip(payload.context, fip,
                                                        fip_request)
        else:
            router_id = old_fip.get('router_id')
            fixed_ip_address = old_fip.get('fixed_ip_address')
            if router_id and fixed_ip_address:
                update_fip = {
                    'id': old_fip['id'],
                    'logical_ip': fixed_ip_address,
                    'external_ip': old_fip['floating_ip_address'],
                    'floating_network_id': old_fip['floating_network_id']
                }
                try:
                    self.l3plugin._ovn_client.disassociate_floatingip(
                        payload.context, update_fip, router_id)
                    self.l3plugin.update_floatingip_status(
                        payload.context, old_fip['id'],
                        constants.FLOATINGIP_STATUS_DOWN)
                except Exception as e:
                    LOG.error('Error in disassociating floatingip %(id)s: '
                              '%(error)s', {'id': old_fip['id'], 'error': e})
        if not fip['router_id']:
            db_rev.delete_revision(payload.context, payload.resource_id,
                                   ovn_const.TYPE_FLOATINGIPS)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_DELETE])
    def _process_floatingip_delete(self, resource, event, trigger, payload):
        if not db_rev.get_revision_row(payload.context, payload.resource_id):
            return
        self.l3plugin._ovn_client.delete_floatingip(payload.context,
                                                    payload.resource_id)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_STATUS_UPDATE])
    def _process_floatingip_status_update(self, resource, event, trigger,
                                          payload):
        if not db_rev.get_revision_row(payload.context, payload.resource_id):
            return
        self.l3plugin._ovn_client.update_floatingip_status(payload.context,
                                                           payload.states[0])



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovs import stream
from unittest import mock

from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.services.bgp import ovn
from neutron.tests import base


class OvnNbIdlTestCase(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        ovn_conf.register_opts()
        ovn_conf.cfg.CONF.set_override(
            'ovn_nb_private_key', 'nb-private-key', 'ovn')
        ovn_conf.cfg.CONF.set_override(
            'ovn_nb_certificate', 'nb-certificate', 'ovn')
        ovn_conf.cfg.CONF.set_override(
            'ovn_nb_ca_cert', 'nb-ca-cert', 'ovn')
        mock.patch(
            'ovsdbapp.backend.ovs_idl.idlutils.get_schema_helper').start()
        mock.patch('ovs.db.idl.Idl.__init__').start()

    def test_init_with_ssl(self):
        """Check the SSL is configured correctly"""
        connection = "ssl:127.0.0.1:6640"
        ovn.OvnNbIdl(connection)
        self.assertEqual('nb-ca-cert', stream.Stream._SSL_ca_cert_file)
        self.assertEqual('nb-certificate', stream.Stream._SSL_certificate_file)
        self.assertEqual('nb-private-key', stream.Stream._SSL_private_key_file)



# Copyright (c) 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as const
from ovsdbapp.backend.ovs_idl import event as idl_event

from neutron.agent.common import ovs_lib
from neutron.agent.ovsdb.native import helpers
from neutron.tests.common.exclusive_resources import port
from neutron.tests.common import net_helpers
from neutron.tests.functional import base


class WaitOvsManagerEvent(idl_event.WaitEvent):
    event_name = 'WaitOvsManagerEvent'

    def __init__(self, manager_target, inactivity_probe=None, event=None):
        table = 'Manager'
        events = (self.ROW_CREATE,) if not event else (event,)
        conditions = (('target', '=', manager_target),)
        super().__init__(events, table, conditions, timeout=10)
        self.inactivity_probe = inactivity_probe

    def match_fn(self, event, row, old):
        if (self.inactivity_probe is None or
                self.inactivity_probe == row.inactivity_probe[0]):
            return True
        return False


class EnableConnectionUriTestCase(base.BaseSudoTestCase):

    def test_add_manager_appends(self):
        ovs = ovs_lib.BaseOVS()
        ovsdb_cfg_connections = []
        manager_connections = []
        manager_removal = []

        for _ in range(5):
            _port = self.useFixture(port.ExclusivePort(
                const.PROTO_NAME_TCP,
                start=net_helpers.OVS_MANAGER_TEST_PORT_FIRST,
                end=net_helpers.OVS_MANAGER_TEST_PORT_LAST)).port
            ovsdb_cfg_connections.append('tcp:127.0.0.1:%s' % _port)
            manager_connections.append('ptcp:%s:127.0.0.1' % _port)

        for index, conn_uri in enumerate(ovsdb_cfg_connections):
            target_event = WaitOvsManagerEvent(manager_connections[index])
            ovs.ovsdb.idl.notify_handler.watch_event(target_event)
            helpers.enable_connection_uri(conn_uri)
            manager_removal.append(ovs.ovsdb.remove_manager(
                manager_connections[index]))
            self.addCleanup(manager_removal[index].execute)
            target_event.wait()
            # This check is redundant, the ``target_event`` ensures the
            # ``Manager`` register with the expected targer is created.
            self.assertIn(manager_connections[index],
                          ovs.ovsdb.get_manager().execute())

        for remove in manager_removal:
            remove.execute()

        for connection in manager_connections:
            self.assertNotIn(connection, ovs.ovsdb.get_manager().execute())

    def test_add_manager_overwrites_existing_manager(self):
        ovs = ovs_lib.BaseOVS()

        _port = self.useFixture(port.ExclusivePort(
            const.PROTO_NAME_TCP,
            start=net_helpers.OVS_MANAGER_TEST_PORT_FIRST,
            end=net_helpers.OVS_MANAGER_TEST_PORT_LAST)).port
        ovsdb_cfg_connection = 'tcp:127.0.0.1:%s' % _port
        manager_connection = 'ptcp:%s:127.0.0.1' % _port

        inactivity_probe = 10
        manager_event = WaitOvsManagerEvent(
            manager_connection, inactivity_probe=inactivity_probe)
        ovs.ovsdb.idl.notify_handler.watch_event(manager_event)
        helpers.enable_connection_uri(ovsdb_cfg_connection,
                                      inactivity_probe=inactivity_probe)
        manager_event.wait()
        self.addCleanup(ovs.ovsdb.remove_manager(manager_connection).execute)
        # First call of enable_connection_uri cretes the manager
        # and the list returned by get_manager contains it:
        my_mans = ovs.ovsdb.get_manager().execute()
        self.assertIn(manager_connection, my_mans)

        # after 2nd call of enable_connection_uri with new value of
        # inactivity_probe will keep the original manager only the
        # inactivity_probe value is set:
        inactivity_probe = 100
        manager_event = WaitOvsManagerEvent(
            manager_connection, inactivity_probe=inactivity_probe,
            event='update')
        ovs.ovsdb.idl.notify_handler.watch_event(manager_event)
        helpers.enable_connection_uri(ovsdb_cfg_connection,
                                      inactivity_probe=inactivity_probe)
        manager_event.wait()
        my_mans = ovs.ovsdb.get_manager().execute()
        self.assertIn(manager_connection, my_mans)
